import sys
from typing import Optional

import click
import requests
from requests.compat import urljoin

from goji.auth import get_credentials, set_credentials
from goji.client import JIRAClient
from goji.config import Configuration
from goji.utils import Datetime


def submit_form(
    session: requests.Session, response: requests.Response, data=None
) -> requests.Response:
    from requests_html import HTML

    html = HTML(url=response.url, html=response.text)

    forms = html.find('form')
    if len(forms) == 0:
        raise Exception('Page does have any forms')

    form = forms[0]
    url = form.attrs['action']
    fields = form.find('input')

    data = data or {}

    for field in fields:
        name = field.attrs['name']

        if name not in data:
            value = field.attrs['value']
            data[name] = value

    response = session.post(urljoin(response.url, url), data=data)
    return response


def check_login(client) -> None:
    response = client.get('myself', allow_redirects=False)

    if response.status_code == 302:
        if sys.version_info.major == 2:
            raise click.ClickException(
                'JIRA instances requires SSO login. goji requires Python 3 to do this. Please upgrade to Python 3'
            )

        # JIRA API may redirect to SSO Authentication if auth fails
        # Manually follow redirect, some SSO requires browser user-agent
        response = client.get(
            response.headers['Location'],
            headers={
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/65.0.3325.181 Safari/537.36'
            },
        )

        auth = client.session.auth

        if (
            '<body onLoad="document.myForm.submit()">' in response.text
            or '<body onLoad="submitForm()">' in response.text
        ):
            # Pretend we're a JavaScript client
            response = submit_form(client.session, response)

        response = submit_form(
            client.session,
            response,
            {'ssousername': auth.username, 'password': auth.password},
        )

        client.save_cookies()

    if response.status_code == 401:
        raise click.ClickException('Incorrect credentials. Try `goji login`.')


@click.group()
@click.option('--profile', default='default')
@click.option('--base-url', envvar='GOJI_BASE_URL')
@click.option('--email', envvar='GOJI_EMAIL', default=None)
@click.option('--password', envvar='GOJI_PASSWORD', default=None)
@click.pass_context
def cli(ctx, profile: str, base_url: str, email: Optional[str], password: Optional[str]) -> None:
    config = Configuration.load()
    p = config.profiles.get(profile.lower())

    if p:
        if not base_url:
            base_url = p.url

        if p.email and not email:
            email = p.email

    if not p and profile != 'default':
        raise click.ClickException(f'Profile {profile} not found')

    if not base_url:
        raise click.ClickException('JIRA base URL is not configured')

    if not ctx.obj:
        if ctx.invoked_subcommand == 'login':
            ctx.obj = base_url
        elif email and password:
            ctx.obj = JIRAClient(base_url, auth=(email, password))
        elif email and not password:
            raise click.ClickException('Password is not configured.')
        elif not email and password:
            raise click.ClickException('Email is not configured.')
        else:
            email, password = get_credentials(base_url)

            if not email or not password:
                raise click.ClickException(
                    'Authentication not configured. Run `goji login`.'
                )

            ctx.obj = JIRAClient(base_url, auth=(email, password))

            if len(ctx.obj.session.cookies) > 0:
                check_login(ctx.obj)


@cli.command('whoami')
@click.pass_obj
def whoami(client: JIRAClient) -> None:
    """View information regarding current user"""
    user = client.get_user()
    click.echo(user)


@click.argument('issue_key')
@cli.command('open')
@click.pass_obj
def open_command(client: JIRAClient, issue_key: str) -> None:
    """Open issue in a web browser"""
    url = urljoin(client.base_url, 'browse/%s' % issue_key)
    click.launch(url)


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def show(client: JIRAClient, issue_key: str) -> None:
    """Print issue contents"""
    issue = client.get_issue(issue_key)
    url = urljoin(client.base_url, 'browse/%s' % issue_key)

    click.echo('\x1b[01;32m-> {issue.key}\x1b[0m'.format(issue=issue))
    click.echo('  {issue.summary}\n'.format(issue=issue))

    if issue.description:
        for line in issue.description.splitlines():
            click.echo('  {}'.format(line))

        click.echo('')

    click.echo('  - Status: {issue.status}'.format(issue=issue))
    click.echo('  - Creator: {issue.creator}'.format(issue=issue))
    click.echo('  - Assigned: {issue.assignee}'.format(issue=issue))
    click.echo('  - URL: {url}'.format(url=url))

    if issue.links:
        click.echo('\n  Related issues:')

        for link in issue.links:
            click.echo('    - {}'.format(link))


@click.argument('user', required=False)
@click.argument('issue_key')
@cli.command()
@click.pass_obj
def assign(client: JIRAClient, issue_key: str, user: Optional[str]) -> None:
    """Assign an issue to a user"""

    if user is None:
        user = client.username

    client.assign(issue_key, user)
    click.echo('{} has been assigned to {}.'.format(issue_key, user))


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def unassign(client: JIRAClient, issue_key: str) -> None:
    """Unassign an issue"""

    client.assign(issue_key, None)
    click.echo('{} has been unassigned.'.format(issue_key))


@click.argument('status', required=False)
@click.argument('issue_key')
@cli.command('change-status')
@click.pass_obj
def change_status(client: JIRAClient, issue_key: str, status: Optional[str]) -> None:
    """Change the status of an issue"""
    click.echo('Fetching possible transitions...')
    transitions = client.get_issue_transitions(issue_key)
    if len(transitions) == 0:
        click.echo('No transitions found for {}'.format(issue_key))
        return

    if status is None:
        for index, transition in enumerate(transitions):
            click.echo('{}: {}'.format(index, transition))
        index = click.prompt('Select a transition', type=int)
        if index < 0 or index >= len(transitions):
            click.echo('No transitions match "{}"'.format(index))
            return
    else:
        index = -1
        for idx, transition in enumerate(transitions):
            if transition.name.lower() == status.lower():
                index = idx
        if index < 0:
            click.echo('No transitions match "{}"'.format(status))
            return

    transition = transitions[index]
    client.change_status(issue_key, transition.id)
    click.echo('Okay, the status for {} is now "{}".'.format(issue_key, transition))


@click.option('--message', '-m', help='Message to comment.')
@click.argument('issue_key')
@cli.command()
@click.pass_obj
def comment(client: JIRAClient, message: Optional[str], issue_key: str) -> None:
    """Comment on an issue"""

    if not message:
        MARKER = '# Leave a comment on {}'.format(issue_key)
        message = click.edit(MARKER)

    if message is None or len(message) == 0:
        return

    try:
        client.comment(issue_key, message)
        click.echo('Comment created')
    except Exception as e:
        click.echo('Comment:\n')
        click.echo(message)
        click.echo('')
        raise e


@click.argument('issue_key')
@cli.command()
@click.pass_obj
def edit(client: JIRAClient, issue_key: str) -> None:
    """Edit issue description"""
    issue = client.get_issue(issue_key)
    description = click.edit(issue.description)

    assert issue.description
    if description is not None and description.strip() != issue.description.strip():
        try:
            client.edit_issue(issue_key, {'description': description.strip()})
            click.echo(
                'Okay, the description for {} has been updated.'.format(issue_key)
            )
        except Exception as e:
            click.echo('There was an issue saving the new description:')
            click.echo(description)
            click.echo('')
            raise e


@click.argument('attachments', type=click.File('rb'), nargs=-1)
@click.argument('issue_key')
@cli.command()
@click.pass_obj
def attach(client: JIRAClient, issue_key: str, attachments) -> None:
    """Attach file(s) to an issue"""

    if len(attachments) == 0:
        click.echo('No attachments to upload')
        return

    for fp in attachments:
        attachments = client.attach(issue_key, fp)

        for attachment in attachments:
            print(f'Attachment {attachment.filename} uploaded.')


@click.argument('summary')
@click.argument('project', metavar='<key>')
@click.option('--type', '-t', required=True)
@click.option('--component', '-c', multiple=True, help='Adds a component')
@click.option('--label', multiple=True, help='Adds a label')
@click.option('--priority', '-p', help='Sets the issue priority')
@click.option('--description', help='Sets the issue description')
@cli.command()
@click.pass_obj
def create(client, project, summary, type, component, label, priority, description):
    """
    Create a new issue, prompting for description contents

    Example:

    \b
        goji create ENG 'Update installation guide' \\
            --type 'Task' \\
            --component Handbook --component External \\
            --priority Low
    """

    if description is None:
        description = click.edit()

    if description is not None:
        fields = {
            'summary': summary,
            'description': description,
            'project': {'key': project},
        }

        if len(component) > 0:
            fields['components'] = [{'name': x} for x in component]

        if len(label) > 0:
            fields['labels'] = label

        if priority is not None:
            fields['priority'] = {'name': priority}

        if type is not None:
            fields['issuetype'] = {'name': type}

        try:
            issue = client.create_issue(fields)
            click.echo('Issue {} created'.format(issue.key))
        except Exception as e:
            click.echo('Description:\n')
            click.echo(description)
            click.echo('')
            raise e


@cli.command()
@click.pass_obj
def login(base_url: str) -> None:
    """Authenticate with JIRA server"""
    email, password = get_credentials(base_url)
    if email is not None:
        if not click.confirm('This server is already configured. Override?'):
            return

    click.echo('Enter your JIRA credentials')

    email = click.prompt('Email', type=str)
    password = click.prompt('Password', type=str, hide_input=True)

    client = JIRAClient(base_url, auth=(email, password))
    check_login(client)
    set_credentials(base_url, email, password)


@click.argument('query')
@click.option('--format', default='{key} {summary}')
@cli.command()
@click.pass_obj
def search(client: JIRAClient, format: str, query: str) -> None:
    """Search issues using JQL"""

    issues = client.search(query)

    for issue in issues:
        format_kwargs = dict(
            key=issue.key,
            summary=issue.summary,
            description=issue.description,
            creator=issue.creator,
            assignee=issue.assignee,
            status=issue.status,
            resolution=issue.resolution,
        )

        format_kwargs.update(issue.customfields)

        click.echo(format.replace('\\n', '\n').format(**format_kwargs))


@cli.group('sprint')
def sprint() -> None:
    pass


@sprint.command('create')
@click.argument('board_id', envvar='GOJI_BOARD_ID', type=int)
@click.argument('name')
@click.option('--start', type=Datetime(format='%d/%m/%y'), default=None)
@click.option('--end', type=Datetime(format='%d/%m/%y'), default=None)
@click.pass_obj
def sprint_create(client: JIRAClient, board_id: int, name: str, start, end) -> None:
    """Create a sprint"""

    client.create_sprint(board_id, name, start_date=start, end_date=end)
    click.echo('Sprint created')



from pathlib import Path
from typing import Dict, Optional

import click
import toml
from jsonschema import ValidationError, validate

SCHEMA = {
  'type': 'object',
  'properties': {
    'profile': { '$ref': '#/definitions/Profiles' },
  },
  'additionalProperties': False,
  'definitions': {
    'Profile': {
      'type': 'object',
      'properties': {
        'url': {
          'type': 'string',
        },
        'email': {
          'type': 'string',
        },
      },
      'required': ['url'],
      'additionalProperties': False,
    },

    'Profiles': {
      'type': 'object',
      'additionalProperties': {
        '$ref': '#/definitions/Profile'
      },
    },
  },
}


class Profile:
    @classmethod
    def from_dict(cls, data) -> 'Profile':
        return cls(data['url'], data.get('email'))

    def __init__(self, url: str, email: Optional[str] = None):
        self.url = url
        self.email = email


class Configuration:
    @classmethod
    def load(cls) -> 'Configuration':
        path = Path.home() / '.config' / 'goji' / 'config.toml'
        if path.exists():
            return cls.fromfile(path)

        return cls({})

    @classmethod
    def fromfile(cls, path: Path) -> 'Configuration':
        with open(path) as fp:
            data = toml.load(fp)

        try:
            validate(instance=data, schema=SCHEMA)
        except ValidationError as exception:
            message = 'Invalid config in {}, at path /{}, {}'.format(
                path,
                '/'.join(exception.path),
                exception.message
            )
            raise click.ClickException(message)

        profiles: Dict[str, Profile] = {}
        for (profile, data) in data.get('profile', {}).items():
            if profile.lower() in profiles:
                raise click.ClickException(f'Profile {profile} defined more than once')

            profiles[profile] = Profile.from_dict(data)

        return cls(profiles=profiles)

    def __init__(self, profiles: Dict[str, Profile]):
        self.profiles = profiles



from typing import Any, Dict, List, Optional


class Model(object):
    pass


class User(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Optional['User']:
        if json:
            return cls(json.get('name') or '', json['displayName'], json.get('emailAddress'))

        return None

    def __init__(self, username: Optional[str], name: str, email: Optional[str] = None):
        self.username = username
        self.name = name
        self.email = email

    def __str__(self) -> str:
        return '{} ({})'.format(self.name, self.username or self.email)


class Issue(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Issue':
        issue = cls(json['key'])

        if 'fields' in json:
            fields = json['fields']
            issue.summary = fields['summary'].rstrip()
            issue.description = fields.get('description')
            issue.creator = User.from_json(fields.get('creator'))
            issue.assignee = User.from_json(fields.get('assignee'))
            issue.status = Status.from_json(fields['status'])

            resolution = fields.get('resolution', None)
            if resolution:
                issue.resolution = Resolution.from_json(resolution)
            else:
                issue.resolution = None

            links = []
            if 'issuelinks' in fields:
                links = [IssueLink.from_json(link) for link in fields['issuelinks']]
            issue.links = links

            issue.customfields = {}

            for (key, value) in fields.items():
                if key.startswith('customfield_'):
                    issue.customfields[key] = value

        return issue

    def __init__(self, key: str):
        self.key = key
        self.summary = None
        self.description = None
        self.creator: Optional[User] = None
        self.assignee: Optional[User] = None
        self.status: Optional['Status'] = None
        self.resolution: Optional[Resolution] = None
        self.links: List['IssueLink'] = []
        self.customfields: Dict[str, Any] = {}

    def __str__(self):
        return self.key


class IssueLinkType(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'IssueLinkType':
        return cls(json['name'], json['inward'], json['outward'])

    def __init__(self, name: str, inward: str, outward: str):
        self.name = name
        self.inward = inward
        self.outward = outward


class IssueLink(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'IssueLink':
        link_type = IssueLinkType.from_json(json['type'])
        issue_link = cls(link_type)

        if 'outwardIssue' in json:
            issue_link.outward_issue = Issue.from_json(json['outwardIssue'])

        if 'inwardIssue' in json:
            issue_link.inward_issue = Issue.from_json(json['inwardIssue'])

        return issue_link

    def __init__(
        self,
        link_type: IssueLinkType,
        inward_issue: Optional[Issue] = None,
        outward_issue: Optional[Issue] = None,
    ):
        self.link_type = link_type
        self.inward_issue = inward_issue
        self.outward_issue = outward_issue

    def __str__(self) -> str:
        if self.outward_issue:
            direction = self.link_type.outward.capitalize()
            issue = self.outward_issue
        elif self.inward_issue:
            direction = self.link_type.inward.capitalize()
            issue = self.inward_issue
        else:
            raise Exception('IssueLink is missing outward or inward issue')

        return '{direction}: {issue.key} ({issue.status})'.format(
            direction=direction, issue=issue
        )


class Transition(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Transition':
        return cls(json['id'], json['name'])

    def __init__(self, identifier: str, name: str):
        self.id = identifier
        self.name = name

    def __str__(self) -> str:
        return self.name


class Comment(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Comment':
        comment = cls(json['id'], json['body'])

        if 'author' in json:
            comment.author = User.from_json(json['author'])

        return comment

    def __init__(self, identifier: str, message: str):
        self.id = identifier
        self.message = message
        self.author: Optional[User] = None

    def __str__(self) -> str:
        return self.message


class Sprint(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Sprint':
        return cls(json['id'], json['name'], json['state'])

    def __init__(self, identifier: str, name: str, state):
        self.id = identifier
        self.name = name
        self.state = state

    def __str__(self) -> str:
        return self.name


class Status(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Status':
        return cls(json['id'], json['name'], json.get('description', None))

    def __init__(self, identifier: str, name: str, description: Optional[str] = None):
        self.id = identifier
        self.name = name
        self.description = description

    def __str__(self) -> str:
        return self.name


class Resolution(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Resolution':
        return cls(json['id'], json['name'], json.get('description', None))

    def __init__(self, identifier: str, name: str, description: Optional[str]):
        self.id = identifier
        self.name = name
        self.description = description

    def __str__(self) -> str:
        return self.name


class Attachment(Model):
    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> 'Attachment':
        return cls(json.get('filename'), int(json['size']))

    def __init__(self, filename: Optional[str], size: int):
        self.filename = filename
        self.size = size


"""
class IssueType(object):
    id
    description
    name

class Issue(object):
    issue type
    creator
    assignee
    url
    summary
    description
"""



import datetime
import mimetypes
import os
import pickle
from typing import Any, List, Optional

import click
import requests
from requests.auth import AuthBase, HTTPBasicAuth
from requests.compat import urljoin

from goji.models import Attachment, Comment, Issue, Sprint, Transition, User


class JIRAException(click.ClickException):
    def __init__(self, error_messages: List[str], errors):
        self.error_messages = error_messages
        self.errors = errors

    def show(self):
        for error in self.error_messages:
            click.echo(error)

        for (key, error) in self.errors.items():
            click.echo('- {}: {}'.format(key, error))


class NoneAuth(AuthBase):
    """
    Creates a "None" auth type as if actual None is set as auth and a netrc
    credentials are found, python-requests will use them instead.
    """

    def __call__(self, request):
        return request


class JIRAAuth(HTTPBasicAuth):
    def __call__(self, request):
        if 'Cookie' in request.headers:
            # Prevent authorization headers when cookies are present as it
            # causes silent authentication errors on the JIRA instance if
            # cookies are used and invalid authorization headers are sent
            # (although request succeeds)

            if 'atlassian.xsrf.token' in request.headers['Cookie'] and len(request.headers['Cookie'].split('=')) == 2:
                # continue if the cookie is ONLY the xsrf token
                # check is very naive as to not get into cookie parsing
                # ensure that we check only for key=value (once) being xsrf
                return super(JIRAAuth, self).__call__(request)

            return request

        return super(JIRAAuth, self).__call__(request)


class JIRAClient(object):
    def __init__(self, base_url: str, auth=None):
        self.session = requests.Session()
        self.base_url = base_url
        self.rest_base_url = urljoin(self.base_url, 'rest/api/2/')

        if auth:
            self.session.auth = JIRAAuth(auth[0], auth[1])
        else:
            self.session.auth = NoneAuth()

        self.load_cookies()

    # Persistent Cookie

    @property
    def cookie_path(self) -> str:
        return os.path.expanduser('~/.goji/cookies')

    def load_cookies(self) -> None:
        if os.path.exists(self.cookie_path):
            try:
                with open(self.cookie_path, 'rb') as fp:
                    self.session.cookies = pickle.load(fp)
            except Exception as e:
                print('warning: Could not load cookies from dist: {}'.format(e))

    def save_cookies(self) -> None:
        cookies = self.session.cookies.keys()
        cookies.remove('atlassian.xsrf.token')

        if len(cookies) > 0:
            os.makedirs(os.path.expanduser('~/.goji'), exist_ok=True)

            with open(self.cookie_path, 'wb') as fp:
                pickle.dump(self.session.cookies, fp)
        elif os.path.exists(self.cookie_path):
            os.remove(self.cookie_path)

    # Methods

    def validate_response(self, response: requests.Response) -> None:
        if response.status_code >= 400 and 'application/json' in response.headers.get(
            'Content-Type', ''
        ):
            error = response.json()
            raise JIRAException(error.get('errorMessages', []), error.get('errors', {}))

    def get(self, path: str, **kwargs) -> requests.Response:
        url = urljoin(self.rest_base_url, path)
        response = self.session.get(url, **kwargs)
        self.validate_response(response)
        return response

    def post(self, path: str, json) -> requests.Response:
        url = urljoin(self.rest_base_url, path)
        response = self.session.post(url, json=json)
        self.validate_response(response)
        return response

    def put(self, path: str, json) -> requests.Response:
        url = urljoin(self.rest_base_url, path)
        response = self.session.put(url, json=json)
        self.validate_response(response)
        return response

    @property
    def username(self) -> Optional[str]:
        if self.session.auth and isinstance(self.session.auth, JIRAAuth):
            return self.session.auth.username

        return None

    def get_user(self) -> Optional[User]:
        response = self.get('myself', allow_redirects=False)
        response.raise_for_status()
        return User.from_json(response.json())

    def get_issue(self, issue_key: str) -> Issue:
        response = self.get('issue/%s' % issue_key)
        response.raise_for_status()
        return Issue.from_json(response.json())

    def get_issue_transitions(self, issue_key: str) -> List[Transition]:
        response = self.get('issue/%s/transitions' % issue_key)
        response.raise_for_status()
        return list(map(Transition.from_json, response.json()['transitions']))

    def change_status(self, issue_key: str, transition_id: str) -> None:
        data = {'transition': {'id': transition_id}}
        self.post('issue/%s/transitions' % issue_key, data)

    def edit_issue(self, issue_key: str, updated_fields) -> None:
        data = {'fields': updated_fields}
        self.put('issue/%s' % issue_key, data)

    def attach(self, issue_key: str, attachment) -> List[Attachment]:
        media_type = mimetypes.guess_type(attachment.name)
        files = {
            'file': (attachment.name, attachment, media_type[0]),
        }
        url = urljoin(self.rest_base_url, f'issue/{issue_key}/attachments')
        response = self.session.post(
            url,
            headers={'X-Atlassian-Token': 'no-check'},
            files=files,
        )
        self.validate_response(response)
        return list(map(Attachment.from_json, response.json()))

    def create_issue(self, fields) -> Issue:
        response = self.post('issue', {'fields': fields})
        return Issue.from_json(response.json())

    def assign(self, issue_key: str, name: Optional[str]) -> None:
        response = self.put('issue/%s/assignee' % issue_key, {'name': name})
        response.raise_for_status()

    def comment(self, issue_key: str, comment: str) -> Comment:
        response = self.post('issue/%s/comment' % issue_key, {'body': comment})
        return Comment.from_json(response.json())

    def search(self, query: str) -> List[Issue]:
        response = self.post('search', {'jql': query})
        response.raise_for_status()
        return list(map(Issue.from_json, response.json()['issues']))

    def create_sprint(
        self,
        board_id: int,
        name: str,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
    ) -> Sprint:
        payload = {
            'originBoardId': board_id,
            'name': name,
        }

        if start_date:
            payload['startDate'] = start_date.isoformat()

        if end_date:
            payload['endDate'] = end_date.isoformat()

        url = urljoin(self.base_url, 'rest/agile/1.0/sprint')
        response = self.session.post(url, json=payload)
        self.validate_response(response)
        return Sprint.from_json(response.json())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['goji']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'requests-html>=0.10.0,<0.11.0',
 'requests[socks]>=2.25.1,<3.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['goji = goji.commands:cli']}

setup_kwargs = {
    'name': 'goji',
    'version': '0.4.0',
    'description': 'Command line JIRA client',
    'long_description': None,
    'author': 'Kyle Fuller',
    'author_email': 'kyle@fuller.li',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
